/***************************************************************************
                    DirectDraw_display.c  -  description
                             -------------------
   Display module using the Direct Draw API.
                             -------------------
    begin                : Mon Jan 24 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <stdlib.h>

#include "win32_main/winmain.h"
#include <wingdi.h>

#include "common/Timer.h"

#define BACK_BUFFERS    1

static const char *four_CC    = "YV12";
static u8         *cur_frame  = NULL,
                  *ref_buf    = NULL;
static u32         buf_size   = 0,
                   ovr_wid    = 0,
                   ovr_hgt    = 0;

static int  init_DirectDrawOverlay();
static void clear_DirectDrawOverlay();


/* Prepares the display for the incoming 4:2:0 video stream.
      INPUT: src_wid  - Width of the source frame.
             src_hgt  - Height of the source frame.
             disp_wid - Width of the displayed frame.
             disp_hgt - Height of the displayed frame.
      RETURN: Success - Any non-zero value.
              Failure - 0 */
int DISPLAY_init(u32 src_wid, u32 src_hgt, u32 disp_wid, u32 disp_hgt)
{
   if ( (win_wid != disp_wid)||(win_hgt != disp_hgt) ) {
      RECT box = {0, 0, 0, 0};

      if ( (!win_wid)||(!win_hgt) ) {
         win_wid = disp_wid;
         win_hgt = disp_hgt;
      }
      else {
         win_wid = disp_wid;
         win_hgt = disp_hgt;
         if (GetClientRect(win, &box)) {
            disp_wid = (u32)box.right;
            disp_hgt = (u32)box.bottom;
         }
      }

      get_win_size(disp_wid, disp_hgt);

      if ( (!box.right)&&(GetClientRect(win, &box)) ) {
         /* resize the window */
         ClientToScreen(win, (POINT *) &(box.left));
         ClientToScreen(win, (POINT *) &(box.right));
         box.left  += (long)win_x;
         box.top   += (long)win_y;
         box.right  = box.left + (long)win_wid;
         box.bottom = box.top  + (long)win_hgt;
         AdjustWindowRectEx(&box, DWSTYLE, MENU_PRESENT, DWEXSTYLE);
         if (box.right > (long)screen_wid) {
            box.left -= box.right - (long)screen_wid;
            box.right = (long)screen_wid;
         }
         if (box.bottom > (long)screen_hgt) {
            box.top   -= box.bottom - (long)screen_hgt;
            box.bottom = (long)screen_hgt;
         }
         MoveWindow(win, (int)box.left, (int)box.top,
                    (int)(box.right - box.left),
                    (int)(box.bottom - box.top),
                    TRUE);
      }
   }

   if ( (ovr_wid != src_wid)||(ovr_hgt != src_hgt) ) {
      DISPLAY_done();
      ovr_wid = src_wid;
      ovr_hgt = src_hgt;
      if (!init_DirectDrawOverlay())
         return 0;   /* could not re-initialize Direct Draw */
   }
   else
      clear_DirectDrawOverlay();
   return 1;
}

/* Returns the buffer that will hold the reference frames. */
u8 *DISPLAY_ref_buf(u32 size)
{
   if (size > buf_size) {
      /* create a bigger buffer */
      if (ref_buf)
         free(ref_buf);
      ref_buf  = malloc((size_t)size);
      buf_size = (ref_buf) ? size : 0;
   }
   return ref_buf;
}

/* Called when the display is no longer needed. */
void DISPLAY_done()
{
   if ( (overlay)&&(surface) ) {
      IDirectDrawSurface_UpdateOverlay(overlay, NULL, surface, NULL,
                                       DDOVER_HIDE, NULL);
      IDirectDrawSurface_Release(overlay);
      overlay = overlay_back = NULL;
      ovr_wid = ovr_hgt = 0;
   }
   if (ref_buf) {
      free(ref_buf);
      ref_buf  = NULL;
      buf_size = 0;
   }
}

/* Returns the width of the frame, this module is expecting.
   This value may not be equal to src_wid, so do not ignore. */
u32 DISPLAY_frame_wid()
{
   if (overlay) {
      DDSURFACEDESC desc;

      memset(&desc, 0, sizeof(DDSURFACEDESC));
      desc.dwSize = sizeof(DDSURFACEDESC);
      IDirectDrawSurface_Lock(overlay, NULL, &desc,
                              (DDLOCK_WAIT | DDLOCK_NOSYSLOCK), NULL);
      IDirectDrawSurface_Unlock(overlay, desc.lpSurface);

      return (u32)desc.lPitch;
   }
   return 0;
}

/* Returns a pointer to the next frame to be displayed.
   Returns NULL if DISPLAY_init was not called, or failed. */
u8 *DISPLAY_get_next_frame()
{
   if ( (!cur_frame)&&(overlay_back) ) {
      DDSURFACEDESC desc;

      memset(&desc, 0, sizeof(DDSURFACEDESC));
      desc.dwSize = sizeof(DDSURFACEDESC);
      IDirectDrawSurface_Lock(overlay_back, NULL, &desc,
                              (DDLOCK_WAIT | DDLOCK_NOSYSLOCK), NULL);
      cur_frame = desc.lpSurface;
   }
   return cur_frame;
}

/* Called when the next frame should be displayed.
   Returns a non-zero value, if the frame is late. */
u32 DISPLAY_next_frame_done(CLK_VAL pts, CLK_VAL offset)
{
#ifndef _SPEED_TEST_
   s32 sleep_time = (s32)(pts - MPEG2_prog_clk(offset));
#endif

   if (cur_frame) {
#ifndef _SPEED_TEST_
# if (MPEG2_SYS_SLOW_DOWN > 1)
      sleep_time *= MPEG2_SYS_SLOW_DOWN;
# endif
      if (sleep_time > (MPEG2_SYS_CLK_FREQ / 1000)) {
         sleep_time /=  MPEG2_SYS_CLK_FREQ / 1000;
         TIMER_sleep(sleep_time);
      }
#endif

      IDirectDrawSurface_Unlock(overlay_back, cur_frame);
      cur_frame = NULL;
      IDirectDrawSurface_Flip(overlay, NULL, DDFLIP_WAIT);
   }
#ifndef _SPEED_TEST_
   return (sleep_time < -MAX_LAG_TIME);
#else
   return 0;
#endif
}


static int init_DirectDrawOverlay()
{
   const char   *err_msg;
   DDSURFACEDESC desc;
   DDSCAPS       caps;

   if ( (!lpDD)||(!surface) ) {
      err_msg = "Direct Draw was not initialized.";
      goto INIT_DD_ERR;
   }

   memset(&desc, 0, sizeof(DDSURFACEDESC));
   desc.dwSize            = sizeof(DDSURFACEDESC);
   desc.dwFlags           = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH |
                            DDSD_PIXELFORMAT | DDSD_BACKBUFFERCOUNT;
   desc.ddsCaps.dwCaps    = DDSCAPS_OVERLAY | DDSCAPS_VIDEOMEMORY |
                            DDSCAPS_FLIP | DDSCAPS_COMPLEX;
   desc.dwWidth           = (DWORD)ovr_wid;
   desc.dwHeight          = (DWORD)ovr_hgt;
   desc.dwBackBufferCount = BACK_BUFFERS;

   desc.ddpfPixelFormat.dwSize        = sizeof(DDPIXELFORMAT);
   desc.ddpfPixelFormat.dwFlags       = DDPF_FOURCC;
   desc.ddpfPixelFormat.dwFourCC      = *((DWORD *)four_CC);
   desc.ddpfPixelFormat.dwYUVBitCount = 16;
   if (IDirectDraw_CreateSurface(lpDD, &desc, &overlay, NULL) != DD_OK) {
      err_msg = "Cannot create a Direct Draw overlay surface.";
      goto INIT_DD_ERR;
   }

   caps.dwCaps = DDSCAPS_BACKBUFFER;
   if (IDirectDrawSurface_GetAttachedSurface(overlay,
                                             &caps,
                                             &overlay_back) != DD_OK) {
      err_msg = "Cannot get the Direct Draw hidden overlay surface.";
      goto INIT_DD_ERR;
   }

   clear_DirectDrawOverlay();
   resize_DirectDrawOverlay();
   return 1;

INIT_DD_ERR:
   DISPLAY_done();
   MessageBox(win, err_msg,
              ERRMSG_TITLE, (MB_OK|MB_ICONEXCLAMATION));
   return 0;
}


static void clear_DirectDrawOverlay()
{
   DDSURFACEDESC desc;
   size_t        i, luma, chroma;

   for (i = 0; i < (BACK_BUFFERS+1); i++) {
      memset(&desc, 0, sizeof(DDSURFACEDESC));
      desc.dwSize = sizeof(DDSURFACEDESC);

      IDirectDrawSurface_Lock(overlay_back, NULL, &desc,
                              (DDLOCK_WAIT | DDLOCK_NOSYSLOCK), NULL);

      if (!i) {
         luma   = (size_t)desc.lPitch * (size_t)ovr_hgt;
         chroma = ((3*luma) >> 1) - luma;
      }

      cur_frame = desc.lpSurface;
      memset(cur_frame, 0, luma);
      memset((cur_frame + luma), 128, chroma);

      IDirectDrawSurface_Unlock(overlay_back, cur_frame);
      IDirectDrawSurface_Flip(overlay, NULL, DDFLIP_WAIT);
   }
   cur_frame = NULL;
}


#undef BACK_BUFFERS
